import calendar
import logging
from datetime import datetime, timedelta
from typing import Optional

import pytz
from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import Count, Q
from django.utils import timezone

User = get_user_model()

logger = logging.getLogger(__name__)


def get_start_end_dates_by_week(year, week, tz):
    d = datetime(year, 1, 1, tzinfo=tz)
    if d.weekday() <= 3:
        d = d - timedelta(d.weekday())
    else:
        d = d + timedelta(7 - d.weekday())
    dlt = timedelta(days=(week - 1) * 7)
    return d + dlt, d + dlt + timedelta(days=6)


def get_scrobble_count_qs(
    year: Optional[int] = None,
    month: Optional[int] = None,
    week: Optional[int] = None,
    day: Optional[int] = None,
    user=None,
    model_str="Track",
) -> dict[str, int]:
    tz = settings.TIME_ZONE
    if user and user.is_authenticated:
        tz = pytz.timezone(user.profile.timezone)

    tz = pytz.utc
    data_model = apps.get_model(app_label='music', model_name='Track')
    if model_str == "Artist":
        data_model = apps.get_model(app_label='music', model_name='Artist')
    if model_str == "Video":
        data_model = apps.get_model(app_label='videos', model_name='Video')
    if model_str == "SportEvent":
        data_model = apps.get_model(
            app_label='sports', model_name='SportEvent'
        )

    if model_str == "Artist":
        base_qs = data_model.objects.filter(
            track__scrobble__user=user,
            track__scrobble__played_to_completion=True,
        )
    else:
        base_qs = data_model.objects.filter(
            scrobble__user=user,
            scrobble__played_to_completion=True,
        )

    # Returna all media items with scrobble count annotated
    if not year:
        return base_qs.annotate(scrobble_count=Count("scrobble")).order_by(
            "-scrobble_count"
        )

    start = datetime(year, 1, 1, tzinfo=tz)
    end = datetime(year, 12, 31, tzinfo=tz)

    if year and day and month:
        logger.debug('Filtering by year, month and day')
        start = datetime(year, month, day, 0, 0, tzinfo=tz)
        end = datetime(year, month, day, 23, 59, tzinfo=tz)
    elif year and week:
        logger.debug('Filtering by year and week')
        start, end = get_start_end_dates_by_week(year, week, tz)
    elif month:
        logger.debug('Filtering by month')
        end_day = calendar.monthrange(year, month)[1]
        start = datetime(year, month, 1, tzinfo=tz)
        end = datetime(year, month, end_day, tzinfo=tz)

    if model_str == "Artist":
        scrobble_date_filter = Q(
            track__scrobble__timestamp__gte=start,
            track__scrobble__timestamp__lte=end,
        )
        qs = (
            base_qs.filter(scrobble_date_filter)
            .annotate(scrobble_count=Count("track__scrobble", distinct=True))
            .order_by("-scrobble_count")
        )
    else:
        scrobble_date_filter = Q(
            scrobble__timestamp__gte=start, scrobble__timestamp__lte=end
        )
        qs = (
            base_qs.filter(scrobble_date_filter)
            .annotate(scrobble_count=Count("scrobble", distinct=True))
            .order_by("-scrobble_count")
        )

    return qs


def build_charts(
    user: "User",
    year: Optional[int] = None,
    month: Optional[int] = None,
    week: Optional[int] = None,
    day: Optional[int] = None,
    model_str="Track",
):
    ChartRecord = apps.get_model(
        app_label='scrobbles', model_name='ChartRecord'
    )
    results = get_scrobble_count_qs(year, month, week, day, user, model_str)
    unique_counts = list(set([result.scrobble_count for result in results]))
    unique_counts.sort(reverse=True)
    ranks = {}
    for rank, count in enumerate(unique_counts, start=1):
        ranks[count] = rank

    chart_records = []
    for result in results:
        chart_record = {
            'year': year,
            'week': week,
            'month': month,
            'day': day,
            'user': user,
        }
        chart_record['rank'] = ranks[result.scrobble_count]
        chart_record['count'] = result.scrobble_count
        if model_str == 'Track':
            chart_record['track'] = result
        if model_str == 'Video':
            chart_record['video'] = result
        if model_str == 'Artist':
            chart_record['artist'] = result
        chart_records.append(ChartRecord(**chart_record))
    ChartRecord.objects.bulk_create(
        chart_records, ignore_conflicts=True, batch_size=500
    )


def build_yesterdays_charts_for_user(user: "User", model_str="Track") -> None:
    """Given a user calculate needed charts."""
    ChartRecord = apps.get_model(
        app_label='scrobbles', model_name='ChartRecord'
    )
    tz = pytz.timezone(settings.TIME_ZONE)
    if user and user.is_authenticated:
        tz = pytz.timezone(user.profile.timezone)
    now = timezone.now().astimezone(tz)
    yesterday = now - timedelta(days=1)
    logger.info(
        f"Generating charts for yesterday ({yesterday.date()}) for {user}"
    )

    # Always build yesterday's chart
    ChartRecord.build(
        user,
        year=yesterday.year,
        month=yesterday.month,
        day=yesterday.day,
        model_str=model_str,
    )
    now_week = now.isocalendar()[1]
    yesterday_week = now.isocalendar()[1]
    if now_week != yesterday_week:
        logger.info(
            f"New weekly charts for {yesterday.year}-{yesterday_week} for {user}"
        )
        ChartRecord.build(
            user,
            year=yesterday.year,
            month=yesterday_week,
            model_str=model_str,
        )
    # If the month has changed, build charts
    if now.month != yesterday.month:
        logger.info(
            f"New monthly charts for {yesterday.year}-{yesterday.month} for {user}"
        )
        ChartRecord.build(
            user,
            year=yesterday.year,
            month=yesterday.month,
            model_str=model_str,
        )
    # If the year has changed, build charts
    if now.year != yesterday.year:
        logger.info(f"New annual charts for {yesterday.year} for {user}")
        ChartRecord.build(user, year=yesterday.year, model_str=model_str)


def build_missing_charts_for_user(user: "User", model_str="Track") -> None:
    """"""
    ChartRecord = apps.get_model(
        app_label='scrobbles', model_name='ChartRecord'
    )
    Scrobble = apps.get_model(app_label='scrobbles', model_name='Scrobble')

    logger.info(f"Generating historical charts for {user}")
    tz = pytz.timezone(settings.TIME_ZONE)
    if user and user.is_authenticated:
        tz = pytz.timezone(user.profile.timezone)
    now = timezone.now().astimezone(tz)

    first_scrobble = (
        Scrobble.objects.filter(user=user, played_to_completion=True)
        .order_by('created')
        .first()
    )

    start_date = first_scrobble.timestamp
    days_since = (now - start_date).days

    for day_num in range(0, days_since):
        build_date = start_date + timedelta(days=day_num)
        logger.info(f"Generating chart batch for {build_date}")
        ChartRecord.build(user=user, year=build_date.year)
        ChartRecord.build(
            user=user, year=build_date.year, week=build_date.isocalendar()[1]
        )
        ChartRecord.build(
            user=user, year=build_date.year, month=build_date.month
        )
        ChartRecord.build(
            user=user,
            year=build_date.year,
            month=build_date.month,
            day=build_date.day,
        )



# Generated by Django 4.1.5 on 2023-03-03 00:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('scrobbles', '0023_alter_audioscrobblertsvimport_options_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='chartrecord',
            name='period_end',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='chartrecord',
            name='period_start',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



from django.db.models import Count
from django.views import generic
from music.models import Album, Artist, Track
from scrobbles.models import ChartRecord
from scrobbles.stats import get_scrobble_count_qs


class TrackListView(generic.ListView):
    model = Track
    paginate_by = 200

    def get_queryset(self):
        return get_scrobble_count_qs(user=self.request.user).order_by(
            "-scrobble_count"
        )


class TrackDetailView(generic.DetailView):
    model = Track
    slug_field = 'uuid'

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        context_data['charts'] = ChartRecord.objects.filter(
            track=self.object, rank__in=[1, 2, 3]
        )
        return context_data


class ArtistListView(generic.ListView):
    model = Artist
    paginate_by = 100

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(scrobble_count=Count('track__scrobble'))
            .order_by("-scrobble_count")
        )

    def get_context_data(self, *, object_list=None, **kwargs):
        context_data = super().get_context_data(
            object_list=object_list, **kwargs
        )
        context_data['view'] = self.request.GET.get('view')
        return context_data


class ArtistDetailView(generic.DetailView):
    model = Artist
    slug_field = 'uuid'

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        artist = context_data['object']
        rank = 1
        tracks_ranked = []
        scrobbles = artist.tracks.first().scrobble_count
        for track in artist.tracks:
            if scrobbles > track.scrobble_count:
                rank += 1
            tracks_ranked.append((rank, track))
            scrobbles = track.scrobble_count

        context_data['tracks_ranked'] = tracks_ranked
        context_data['charts'] = ChartRecord.objects.filter(
            artist=self.object, rank__in=[1, 2, 3]
        )
        return context_data


class AlbumListView(generic.ListView):
    model = Album

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(scrobble_count=Count('track__scrobble'))
            .order_by("-scrobble_count")
        )


class AlbumDetailView(generic.DetailView):
    model = Album
    slug_field = 'uuid'

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        # context_data['charts'] = ChartRecord.objects.filter(
        #    track__album=self.object, rank__in=[1, 2, 3]
        # )
        return context_data



import pytz
from django.conf import settings
from django.utils import timezone
import calendar

from datetime import datetime, timedelta

# need to translate to a non-naive timezone, even if timezone == settings.TIME_ZONE, so we can compare two dates
def to_user_timezone(date, profile):
    timezone = profile.timezone if profile.timezone else settings.TIME_ZONE
    return date.astimezone(pytz.timezone(timezone))


def to_system_timezone(date):
    return date.astimezone(pytz.timezone(settings.TIME_ZONE))


def now_user_timezone(profile):
    timezone.activate(pytz.timezone(profile.timezone))
    return timezone.localtime(timezone.now())


def start_of_day(dt, profile) -> datetime:
    """Get the start of the day in the profile's timezone"""
    timezone = profile.timezone if profile.timezone else settings.TIME_ZONE
    tzinfo = pytz.timezone(timezone)
    return datetime.combine(dt, datetime.min.time(), tzinfo)


def end_of_day(dt, profile) -> datetime:
    """Get the start of the day in the profile's timezone"""
    timezone = profile.timezone if profile.timezone else settings.TIME_ZONE
    tzinfo = pytz.timezone(timezone)
    return datetime.combine(dt, datetime.max.time(), tzinfo)


def start_of_week(dt, profile) -> datetime:
    # TODO allow profile to set start of week
    return start_of_day(dt, profile) - timedelta(dt.weekday())


def end_of_week(dt, profile) -> datetime:
    # TODO allow profile to set start of week
    return start_of_week(dt, profile) + timedelta(days=6)


def start_of_month(dt, profile) -> datetime:
    return start_of_day(dt, profile).replace(day=1)


def end_of_month(dt, profile) -> datetime:
    next_month = end_of_day(dt, profile).replace(day=28) + timedelta(days=4)
    # subtracting the number of the current day brings us back one month
    return next_month - timedelta(days=next_month.day)


def start_of_year(dt, profile) -> datetime:
    return start_of_day(dt, profile).replace(month=1, day=1)



# Generated by Django 4.1.5 on 2023-03-02 19:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('music', '0012_album_allmusic_id_album_discogs_id_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='album',
            name='theaudiodb_score',
            field=models.FloatField(blank=True, null=True),
        ),
    ]
